#include <omp.h>
#include <stdio.h>
#include <stdlib.h>

#define UB 2000
// #define BS 40

// define a min function
#define min(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
       _a < _b ? _a : _b; })

int main (int argc, char *argv[]) {


   int BS = 40;
   int bs[6] = {20, 40, 100, 200, 400, 1000};
   

   // allocate 2 dimensional arrays for A, B and C
   double **a = malloc(UB * sizeof(double *));
   printf("**a value %p\n", a); fflush(stdout);
   double **b = malloc(UB * sizeof(double *));
   printf("**b value %p\n", b); fflush(stdout);
   double **c = malloc(UB * sizeof(double *));
   printf("**c value %p\n", c); fflush(stdout);
   a[0] = malloc(UB*UB*sizeof(double));
   printf("a[0] value %p\n", a[0]); fflush(stdout);
   b[0] = malloc(UB*UB*sizeof(double));
   printf("b[0] value %p\n", b[0]); fflush(stdout);
   c[0] = malloc(UB*UB*sizeof(double));
   printf("c[0] value %p\n", c[0]); fflush(stdout);
   for (int i = 1; i < UB; i++) {
      a[i] = a[0] + (UB * i);
      b[i] = b[0] + (UB * i);
      c[i] = c[0] + (UB * i);
   }
   printf("ptr assign finished\n"); fflush(stdout); 

   // initialize the arrays
   for (int i = 0; i < UB; i++) {
      for (int j = 0; j < UB; j++) {
         a[i][j] = 1.0;
         b[i][j] = 1.0;
         c[i][j] = 0.0;
      }
   }
   printf("init finished\n"); fflush(stdout);

   // Perform the unoptimized matrix multiply
   double t = omp_get_wtime( );
   for (int i = 0; i < UB; i++) {
      for (int j = 0; j < UB; j++) {
         for (int k = 0; k < UB; k++) {
            c[i][j] += a[i][k] + b[k][j];
         }
      }
   }
   t = omp_get_wtime( ) - t;
   printf("time for unoptimized MM1 is %g\n", t);

   // Perform the optimized matrix multiply
   t = omp_get_wtime( );
   for (int kk = 0; kk < UB; kk += BS) {
      for(int jj = 0; jj < UB; jj += BS) {
         for(int i = 0; i < UB; i++) {
            for(int k = kk; k <  min(UB, kk+BS); k++) {
               for(int j = jj; j < min(UB, jj + BS); j++) {
                  c[i][j] += a[i][k]*b[k][j];
               }
            }
         }
      }
   }
   // get the elapsed time and print it.
   t = omp_get_wtime( ) - t;
   printf("time for optimized MM1 is %g\n", t);

   for (int m = 0; m < 6; m++) { 
      BS = bs[m];
      // Perform the optimized matrix multiply
      t = omp_get_wtime( );
      for (int kk = 0; kk < UB; kk += BS) {
         for(int jj = 0; jj < UB; jj += BS) {
         #pragma omp parallel for 
            for(int i = 0; i < UB; i++) {
               for(int k = kk; k <  min(UB, kk+BS); k++) {
                  for(int j = jj; j < min(UB, jj + BS); j++) {
                     c[i][j] += a[i][k]*b[k][j];
                  }
               }
            }
         }
      }
      // get the elapsed time and print it.
      t = omp_get_wtime( ) - t;
      printf("time for parallel optimized MM1 with bs %d is %g\n", BS, t);
   }
}
            
 

   


